/*global print, File */

var RESULT_DIRECTORY = 'results';
var HTML_DIRECTORY = 'html';

/**
 * An object which holds parameters for a cell in a table
 * @param {String}  inputText The text the cell should contain
 * @param {Boolean} header    Whether or not the cell is in a header
 */
function Cell(inputText, header) {
    'use strict';
    this.text = '';
    if (typeof inputText !== 'undefined') {
        this.text = inputText;
    }

    this.isHeader = false;
    if (typeof header !== 'undefined') {
        this.isHeader = header;
    }

    this.bold = false;
    this.color = '#000000';
    this.align = 'left';
    this.link = null;
}

/**
 * An object which holds information about a table
 * @param {Number}  numColumns The number of columns in the table
 * @param {Boolean} noHeader   Whether or not the table has a header
 */
function Table(numColumns, noHeader) {
    'use strict';
    this.columns = numColumns;

    var startWithHeader = true;
    if (typeof noHeader !== 'undefined') {
        startWithHeader = !noHeader;
    }

    this.rows = [];

    /**
     * Adds rows to this Table
     * @param {Number}  rowsToAdd The number of rows to add
     * @param {Boolean} header    Whether the row is/rows are considered a header
     * @param {String}  text      The default text to set each cell to in the row(s)
     */
    this.addRows = function (rowsToAdd, header, text) {
        var newRow = [],
            i,
            j,
            isHeader = false,
            defaultText = '';

        if (typeof header !== 'undefined') {
            isHeader = header;
        }

        if (typeof text !== 'undefined') {
            defaultText = text;
        }

        for (i = 0; i < rowsToAdd; i = i + 1) {
            for (j = 0; j < this.columns; j = j + 1) {
                newRow.push(new Cell(defaultText, isHeader));
            }
            this.rows.push(newRow);
        }
    };

    if (startWithHeader) {
        this.addRows(1, true);
    }
}

/**
 * An object containing information needed for a page that holds information about errors and assertions
 * @param {String}  caseName        The name of the test case this page is for
 * @param {Boolean} isAssertionPage Whether the page is for assertions or failures
 */
function ErrorsAssertionsPage(caseName, isAssertionPage) {
    'use strict';
    this.name = '';
    if (isAssertionPage) {
        this.name = caseName + ' Assertions';
    } else {
        this.name = caseName + ' Failures';
    }
    this.errorList = [];
    this.assertionsPage = isAssertionPage;
}

/**
 * An object which holds information about a test case
 * @param   {String}   caseName The name of the test case
 * @param   {String}   desc     The test case's description
 */
function TestCasePage(caseName, desc) {
    'use strict';
    var i,
        headerRow,
        currentCell,
        currentRow;
    this.name = caseName;
    this.description = 'No Description Provided';
    if (typeof desc !== 'undefined') {
        this.description = desc;
    }
    this.errorPage = new ErrorsAssertionsPage(this.name, false);
    this.assertionsPage = new ErrorsAssertionsPage(this.name, true);

    this.lastError = null;
    this.lastAssertion = null;

    this.numAssertions = 0;
    this.maxPrintAssertions = 100;
    this.hitMax = false;

    this.testFailed = false;
    this.resultTable = new Table(6);
    /** set up the headers */
    headerRow = this.resultTable.rows[0];
    headerRow[0].text = 'Time Started';
    headerRow[1].text = 'Time Finished';
    headerRow[2].text = 'Result';
    headerRow[3].text = 'Time (seconds)';
    headerRow[4].text = 'Iterations';
    headerRow[5].text = 'Average Time (seconds)';

    /** set up the output row */
    this.resultTable.addRows(1);
    for (i = 0; i < this.resultTable.columns; i = i + 1) {
        this.resultTable.rows[1][i].align = 'center';
    }

    /**
     * Logs a test failure into this object
     * @param {String} errorString The test failure information
     */
    this.logError = function (errorString) {
        this.lastError = errorString;
        this.errorPage.errorList.push(errorString);
    };

    /**
     * Logs an assertion into this object
     * @param {String} assertionString The assertion information
     */
    this.logAssertion = function (assertionString) {
        this.numAssertions++;
        if(this.numAssertions < this.maxPrintAssertions) {
            print(assertionString);
        } else {
            if(!this.hitMax) {
                this.hitMax = true;
                print('Logged more than ' + this.maxPrintAssertions + ' assertions, no more for this case will be printed, but will still be logged.');
            }
        }

        this.lastAssertion = assertionString;
        this.assertionsPage.errorList.push(assertionString);
    };

    /************************ set output table values *************************/

    this.testOutputSet = false;

    this.testPassed = null;
    this.testTime = null;
    this.testIterations = null;

    /**
     * Sets up the test case's output table
     * @param   {Object}   startedDate  A date object containing the time the test case was started
     * @param   {Object}   finishedDate A date object containing the time the test case finished
     * @param   {Boolean}  passed       Whether or not the test case passed
     * @param   {Number}   time         The time, in seconds, it took for the test case to finish
     * @param   {Number}   iterations   The number of iterations the test case ran
     */
    this.setOutput = function (startedDate, finishedDate, passed, time, iterations) {
        this.outputSet = true;

        this.testPassed = passed;
        this.testTime = time;
        this.testIterations = iterations;

        /**
         * A miniature function that returns the current date and time as a string
         * in the format yyyy-mm-dd hh:mm:ss
         * @param   {Object} date A date object
         * @returns {String} The date as a string
         */
        function dateToTimeString(date) {
            return date.getUTCFullYear() + '-' + (date.getMonth() + 1) + '-' +
                date.getUTCDate() + ' ' + date.getHours() + ':' +
                date.getMinutes() + ':' + date.getSeconds();
        }

        currentRow = this.resultTable.rows[1];

        /** startedDate */
        currentCell = currentRow[0];
        currentCell.text = dateToTimeString(startedDate);

        /** finishedDate */
        currentCell = currentRow[1];
        currentCell.text = dateToTimeString(finishedDate);

        /** passed */
        currentCell = currentRow[2];
        currentCell.bold = true;
        if (passed) {
            currentCell.text = 'Pass';
            currentCell.color = '#008000';
        } else {
            currentCell.text = 'Fail';
            currentCell.color = '#A52A2A';
            this.testFailed = true;
        }

        /** time */
        currentCell = currentRow[3];
        currentCell.text = time.toString();

        /** iterations */
        currentCell = currentRow[4];
        currentCell.text = iterations.toString();

        /** average time */
        currentCell = currentRow[5];
        currentCell.text = (time / iterations).toFixed(3);
    };
}

/**
 * An object which holds information about a test suite
 * @param {String} suiteName The test suite's name
 * @param {String} desc      The test suite's description
 */
function TestSuitePage(suiteName, desc) {
    'use strict';
    var i,
        headerRow,
        ttHeaderRow,
        numTests = 0,
        numPassed = 0,
        numFailed = 0,
        totalTime = 0;
    this.name = suiteName;
    this.description = 'No Description Provided';
    if (typeof desc !== 'undefined') {
        this.description = desc;
    }

    this.metricsTable = new Table(4);

    /** set up the headers */
    headerRow = this.metricsTable.rows[0];
    headerRow[0].text = 'Total Tests';
    headerRow[1].text = 'Tests Passed';
    headerRow[2].text = 'Tests Failed';
    headerRow[3].text = 'Total Time (seconds)';

    /** set up the output row */
    this.metricsTable.addRows(1);
    for (i = 0; i < this.metricsTable.columns; i = i + 1) {
        this.metricsTable.rows[1][i].align = 'center';
    }
    
    this.getTotalTime = function() {
        return totalTime;
    };

    this.testsTable = new Table(5);

    /** set up the headers */
    ttHeaderRow = this.testsTable.rows[0];
    ttHeaderRow[0].text = 'Name';
    ttHeaderRow[1].text = 'Description';
    ttHeaderRow[2].text = 'Result';
    ttHeaderRow[3].text = 'Time (seconds)';
    ttHeaderRow[4].text = 'Iterations';

    /**
     * A function which adds a row to the current test suite which contains
     * information about a test case
     * @param {Object} testCase The test case to create a row for
     */
    this.addTestRow = function (testCase) {
        numTests = numTests + 1;
        var testName = testCase.name,
            testDesc = testCase.description,
            testPassed = testCase.testPassed,
            testTime = testCase.testTime,
            testIterations = testCase.testIterations,
            currentRow,
            currentCell;

        this.testsTable.addRows(1);

        currentRow = this.testsTable.rows[this.testsTable.rows.length - 1];

        /** Name Column */
        currentCell = currentRow[0];
        currentCell.text = testName;
        currentCell.link = HTML_DIRECTORY + '/' + testName;

        /** Description Column */
        currentCell = currentRow[1];
        currentCell.text = testDesc;

        /** Result Column */
        currentCell = currentRow[2];
        currentCell.bold = true;
        currentCell.align = 'center';
        if (testPassed) {
            currentCell.text = 'Pass';
            currentCell.color = '#008000';
            numPassed = numPassed + 1;
        } else {
            currentCell.text = 'Fail';
            currentCell.color = '#A52A2A';
            numFailed = numFailed + 1;
        }

        /** Time Column */
        currentCell = currentRow[3];
        currentCell.align = 'center';
        currentCell.text = testTime.toString();
        totalTime += testTime;

        /** Iterations Column */
        currentCell = currentRow[4];
        currentCell.align = 'center';
        currentCell.text = testIterations.toString();
    };

    /**
     * A function which fills out the test suite's metrics table with information
     */
    this.setMetrics = function () {
        var metricsRow = this.metricsTable.rows[1];

        metricsRow[0].text = numTests.toString();
        metricsRow[1].text = numPassed.toString();
        metricsRow[2].text = numFailed.toString();
        metricsRow[3].text = totalTime.toFixed(3);
    };
}

/**
 * A function which converts a Cell object to its HTML representation
 * @param   {Object} cell The cell object to convert to HTML
 * @returns {String} The HTML representation of the cell
 */
function cellToHTML(cell) {
    'use strict';
    var text = cell.text,
        style = '';

    if (cell.isHeader) {
        return '<th>' + text + '</th>';
    }

    if (cell.bold) {
        text = '<strong>' + text + '</strong>';
    }

    if (cell.link) {
        text = '<a href="' + cell.link + '.html">' + text + '</a>';
    }

    style = 'style="';
    style += 'text-align: ' + cell.align + ';';
    style += 'color:' + cell.color + ';';
    style += '"';

    return '<td ' + style + '>' + text + '</td>';
}

/**
 * A function which converts a row in a table to HTML
 * @param   {Array}  row The row, or array of cells to convert
 * @returns {String} The HTML representation of the row
 */
function rowToHTML(row) {
    'use strict';
    var cells = '',
        i;

    for (i = 0; i < row.length; i = i + 1) {
        cells += cellToHTML(row[i]);
    }

    return '<tr>' + cells + '</tr>';
}

/**
 * A function which returns the HTML equivalent of a table object
 * @param   {Object} table The table to convert
 * @returns {String} The HTML representation of the table
 */
function tableToHTML(table) {
    'use strict';
    var rows = '',
        tbody = '',
        i;

    for (i = 0; i < table.rows.length; i = i + 1) {
        rows += rowToHTML(table.rows[i]);
    }

    tbody = '<tbody>' + rows + '</tbody>';
    return '<table border="1">' + tbody + '</table>';
}

/**
 * A function which converts a TestSuitePage to HTML
 * @param   {Object} suitePage A TestSuitePage which is ready to be converted to HTML
 * @param   {String} font      The font family to use
 * @returns {String} HTML which contains all of the information about a test suite
 */
function suitePageToHTML(suitePage, font) {
    'use strict';
    var suiteFont = 'arial,helvetica,sans-serif',
        title,
        head,
        boldTitle,
        description,
        metrics,
        tests,
        body;

    if (typeof font !== 'undefined') {
        suiteFont = font;
    }
    suitePage.setMetrics();

    title = '<title>' + suitePage.name + '</title>';
    head = ('<head>' + title + '</head>');
    boldTitle = '<h2>' + suitePage.name + '</h2>';
    description = '<p><em>' + suitePage.description + '</em></p>';
    metrics = tableToHTML(suitePage.metricsTable);
    tests = tableToHTML(suitePage.testsTable);
    body = '<body style="font-family:' + suiteFont + ';">' + boldTitle + description + metrics + tests + '</body>';

    return '<html>' + head + body + '</html>';

}

/**
 * A function which returns the HTML equivalent of a TestCasePage
 * @param   {Object} casePage A TestCasePage object which is ready to be converted to HTML
 * @param   {String} font     The font family to use
 * @returns {String} The HTML representation of the TestCasePage
 */
function casePageToHTML(casePage, font) {
    'use strict';
    var caseFont = 'arial,helvetica,sans-serif',
        title,
        head,
        boldTitle,
        description,
        resultTable,
        body;

    if (typeof font !== 'undefined') {
        caseFont = font;
    }
    title = '<title>' + casePage.name + '</title>';
    head = ('<head>' + title + '</head>');
    boldTitle = '<h2>' + casePage.name + '</h2>';
    description = '<p><em>' + casePage.description + '</em></p>';
    resultTable = tableToHTML(casePage.resultTable);
    body = '<body style="font-family:' + caseFont + ';">' + boldTitle + description + resultTable;

    if (casePage.lastError) {
        body += '<h3>Failure Information:</h3>';
        body += '<blockquote><p><pre style="font-family:courier new,courier,monospace;">' + casePage.lastError + '</p></pre></blockquote>';
        body += '<p><a href="' + casePage.name + ' Failures.html">Full Failures List</a></p>';

        if (casePage.lastAssertion) {
            body += '<h3>Last Failed Assertion:</h3>';
            body += '<blockquote><p><pre style="font-family:courier new,courier,monospace;">' + casePage.lastAssertion + '</p></pre></blockquote>';
            body += '<p><a href="' + casePage.name + ' Assertions.html">Full Assertions List</a></p>';
        }
    }

    body += '</body>';

    return '<html>' + head + body + '</html>';
}

/**
 * A function which converts an ErrorsAssertionsPage to HTML
 * @param   {Object} errorsAssertionsPage An ErrorsAssertionsPage which is ready to be converted to HTML
 * @param   {String} caseName             The name of the test case that this page is for
 * @param   {String} font                 The font to use for the page's title
 * @returns {String} The HTML representation of the page
 */
function errorsAssertionsPageToHTML(errorsAssertionsPage, caseName, font) {
    'use strict';
    var loggerFont = 'arial,helvetica,sans-serif',
        title,
        head,
        headerText,
        header,
        errors,
        body,
        i;
    
    if (typeof font !== 'undefined') {
        loggerFont = font;
    }
    title = '<title>' + errorsAssertionsPage.name + '</title>';
    head = ('<head>' + title + '</head>');
    headerText = '';

    if (errorsAssertionsPage.assertionsPage) {
        headerText = 'Failed Assertions For';
    } else {
        headerText = 'Failures For';
    }

    header = '<h2><span style="font-family:' + loggerFont + ';">' + headerText + '&nbsp;<a href="' + caseName + '.html">' + caseName + '</a></span></h2>';

    errors = '';

    for (i = 0; i < errorsAssertionsPage.errorList.length; i = i + 1) {
        errors += '<hr /><p>' + errorsAssertionsPage.errorList[i] + '</p>';
    }

    body = '<body><blockquote><pre style="font-family:courier new,courier,monospace;">' + header + errors + '</pre></blockquote></body>';

    return '<html>' + head + body + '</html>';
}

/*exported Html */

/**
 * An object which contains all of the information required to log
 * information about a test suite and output HTML files containing
 * all of that information
 * @param {String} font The font family to use in the HTML files
 */
function Html(font) {
    'use strict';
    this.fontFamily = 'arial,helvetica,sans-serif';

    if (typeof font !== 'undefined') {
        if(font.constructor !== String) {
            throw new Error('Font is optional, but it must be a String if it is provided');
        }
        this.fontFamily = font;
    }

    this.currentSuite = null;
    this.currentCase = null;
    this.cases = [];
    this.passed = true;

    /**
     * Start a new suite in this logger
     * @param {String} suiteName The name of the test suite
     * @param {String} desc      The description of the test suite
     */
    this.startSuite = function (suiteName, desc, tests) {
        this.passed = true;
        if (this.currentSuite) {
            throw new Error('A suite already exists in this logger');
        }
        
        if (suiteName.constructor !== String) {
            throw new Error('This function requires Strings as arguments');
        }
        print('\n************************************************************\nStarting Suite "' + suiteName + 
              '" with ' + tests + ' tests . . . \n************************************************************');
        this.currentSuite = new TestSuitePage(suiteName, desc);
        
    };

    /**
     * Start a new test case in this logger
     * @param {String} caseName The test case's name
     * @param {String} desc     The test case's description
     */
    this.startTestCase = function (caseName, desc, iterations) {
        if (typeof caseName === 'undefined') {  // If arguments are missing
            throw new Error('Missing Required Argument');
        } else if (caseName.constructor !== String) {   // If the test case name is not a string
            throw new Error('This function requires Strings as arguments');
        } else if (!this.currentSuite) {    // If there is no suite to add the test case to
            throw new Error('No suite to add this test to!');
        } else if (this.currentCase) {
            throw new Error('Current case not closed before starting another one'); // If an unfinished test case is already in the suite
        } else if (typeof desc !== 'undefined') {
            if(desc.constructor !== String) {
                throw new Error('Description is optional, but it must be a String if it is provided');
            }
        } 
        print('\nStarting Test "' + caseName + '" with ' + iterations + ' iterations . . . ');
        this.currentCase = new TestCasePage(caseName, desc);
        this.cases.push(this.currentCase);
        
    };

    /**
     * Ends the current test case in this logger
     * @param {Object}  startedDate  A date object containing the time the test case was started
     * @param {Object}  finishedDate A date object containing the time the test case finished
     * @param {Boolean} passed       Whether or not the test case passed
     * @param {Number}  time         The time it took for the test case to run
     * @param {Number}  iterations   The number of different iterations the test case ran
     */
    this.endTestCase = function (startedDate, finishedDate, passed, time, iterations) {
        if(!this.currentCase) {
            throw new Error('Attempted to end a test case when there was no current test case');   
        }
        if (
            (typeof startedDate === 'undefined') ||
                (typeof finishedDate === 'undefined') ||
                (typeof passed === 'undefined') ||
                (typeof time === 'undefined') ||
                (typeof iterations === 'undefined')
        ) {
            throw new Error('Missing Required Arguments');
        }
        
        if (startedDate.constructor !== Date) {
            throw new Error('The start date variable must be a Date');
        } else if (finishedDate.constructor !== Date) {
            throw new Error('The finished date variable must be a Date');
        } else if (passed.constructor !== Boolean) {
            throw new Error('The test passed variable must be a Boolean');
        } else if (time.constructor !== Number) {
            throw new Error('The test time variable must be a Number');
        } else if (iterations.constructor !== Number) {
            throw new Error('The iterations variable must be a Number');
        }

        print('Test Case "' + this.currentCase.name + '" finished ' + iterations +
            ' iterations in ' + time + ' seconds. Passed: ' + passed);

        this.currentCase.setOutput(startedDate, finishedDate, passed, time, iterations);
        this.currentSuite.addTestRow(this.currentCase);
        if (!this.currentCase.testPassed) {
            this.passed = false;
        }
        this.currentCase = null;    // Clears out the current test case in this logger
    };
    
    /**
     * Ends the current test suite and outputs its HTML
     */
    this.endSuite = function () {
        var i;
        if (!this.currentSuite) {
            throw new Error('Attempted to end a suite when one does not exist');
        }

        if(this.cases.length < 1) {
            throw new Error('Attempted to end a suite when there were never any test cases in it');
        }
        
        for (i = 0; i < this.cases.length; i = i + 1) {
            this.writeErrorsAssertionsPageToFile(this.cases[i].errorPage, this.cases[i].name);
            this.writeErrorsAssertionsPageToFile(this.cases[i].assertionsPage, this.cases[i].name);
        }
        for (i = 0; i < this.cases.length; i = i + 1) {
            this.writeCaseToFile(this.cases[i]);
        }
        this.writeSuiteToFile();
        var passedText;
        if(this.passed) {
            passedText = 'passed all tests';
        } else {
            passedText = 'did not pass all tests';
        }
        print('\n** Suite "' + this.currentSuite.name + '" finished in ' + this.currentSuite.getTotalTime().toFixed(2) + ' seconds, and ' + passedText + '. **');
        this.currentSuite = null;
        this.cases = [];
    };

    /**
     * Writes the current test suite to an HTML file
     */
    this.writeSuiteToFile = function () {
        if (!this.currentSuite) {
            throw new Error('No suite to output!');
        }
        var fileName = '';
        if(this.passed) {
            fileName = RESULT_DIRECTORY + '/' + this.currentSuite.name + '.html';
        } else {
            fileName = RESULT_DIRECTORY + '/' + this.currentSuite.name + ' (FAILED)' + '.html';
        }
        var outputFile = new File(fileName, true);
        outputFile.write(suitePageToHTML(this.currentSuite, this.fontFamily), this.passed);
    };

    /**
     * Writes a test case to an HTML file
     * @param {Object} casePage A TestCasePage object
     */
    this.writeCaseToFile = function (casePage) {
        if (!casePage.outputSet) {
            throw new Error('Attempted to write a test case that has not had its output set');
        }
        var outputFile = new File(RESULT_DIRECTORY + '/' + HTML_DIRECTORY + '/' + casePage.name + '.html', true);
        outputFile.write(casePageToHTML(casePage, this.fontFamily));
    };

    /**
     * Writes an ErrorsAssertionsPage object to an HTML file
     * @param {Object} errorsAssertionsPage An ErrorsAssertionsPage object which is ready to be saved to HTML
     * @param {String} caseName             The name of the test case this page is for
     */
    this.writeErrorsAssertionsPageToFile = function (errorsAssertionsPage, caseName) {
        if (errorsAssertionsPage.errorList.length > 0) {
            if (this.passed) {
                throw new Error('Test "' + caseName + '" passed, but still has logged assertions or errors');
            }
            var outputFile = new File(RESULT_DIRECTORY + '/' + HTML_DIRECTORY + '/' + errorsAssertionsPage.name + '.html', true);
            outputFile.write(errorsAssertionsPageToHTML(errorsAssertionsPage, caseName, this.fontFamily));
        }
    };

    /**
     * Logs a test failure for the current test suite
     * @param {String} errorString A string containing information about the failure
     */
    this.logError = function (errorString) {
        if(errorString.constructor !== String) {
            throw new Error('The logger can only log strings');
        }
        if(this.currentCase) {
            this.currentCase.logError(errorString);
        } else {
            throw new Error('No test case to log this failure to');
        }
    };

    /**
     * Logs a failed assertion for the current test suite
     * @param {String} errorString A string containing useful information about the failed assertion
     */
    this.logAssertion = function (errorString) {
        if(errorString.constructor !== String) {
            throw new Error('The logger can only log strings');
        }
        if(this.currentCase) {
            this.currentCase.logAssertion(errorString);
        } else {
            throw new Error('No test case to log this assertion to');
        }
    };
}
