/*exported TestCase */

/**
 * Test Case Object
 * @param   {String}   name        The test's name
 * @param   {Function} func        The function used for the test
 * @param   {Array}    variables   The variables list used for the test
 * @param   {Function} compare     The funciton used to compare the result with the expected outputs
 * @param   {Object}   expect      The expected outputs for each test
 * @param   {String}   description The test's description
 */
function TestCase(name, func, variables, compare, expect, description) {
    if(name.constructor !== String) {
       throw new Error('Name must be a String'); 
    } else if(func.constructor !== Function) {
       throw new Error('Test must be a function'); 
    } else if(variables.constructor !== Array) {
       throw new Error('Variables must be an Array'); 
    } else if(compare.constructor !== Function) {
       throw new Error('Comparison must be a function'); 
    } 
    
    if(typeof description !== 'undefined') {
        if(description.constructor !== String) {
            throw new Error('Description must be a String'); 
        }
    }
    
    var expected = expect;
    var comparison = compare;

    /**
     * Returns the result of the assertions between two variables
     * @param   {Object}   assertResult   The result to compare against
     * @param   {Function} assertComp     The comparison function to use
     * @param   {Object}   assertExpected The expected value to compare
     * @returns {Boolean}  The result of the comparison
     */
    this.assert = function (assertResult, assertComp, assertExpected) {
        if(assertComp.constructor !== Function) {
            throw new Error('Comparison must be a function'); 
        } 
        return assertComp(assertResult, assertExpected);
    };

    /**
     * Sets the expected test result
     * @param {Object} newExpect The expected result of the test
     */
    this.setExpect = function (newExpect) {
        expected = newExpect;
    };

    /**
     * Sets the comparison function used
     * @param {Function} newCompare The new comparison function
     */
    this.setComparison = function (newCompare) {
        if(newCompare.constructor !== Function) {
            throw new Error('Comparison must be a function'); 
        } 
        comparison = newCompare;
    };

    /**
     * Run the test suite's function
     * @param   {Array} argv Variable
     * @returns {Object} The result of the function
     */
    this.run = function (argv) {
        return func.apply(this, argv);
    };
    
    /**
     * Return the test case's description
     * @returns  {String} The description to return
     */
    this.getDescription = function () {
        return description;
    };
    
    /**
     * Return the test case's name
     * @returns  {String} The name to return
     */
    this.getName = function () {
        return name;
    };

    /**
     * Return the test case's variables
     * @returns  {Array} The variables to return
     */
    this.getVariables = function () {
        return variables;
    };

    /**
     * Return the test case's comparison function
     * @returns  {Function} The returned comparison function
     */
    this.getComparison = function () {
        return comparison;
    };

    /**
     * Gets the expected test result
     * @returns {Object} The expected test result
     */
    this.getExpected = function () {
        return expected;
    };
}
