/* exported TestSuite */
/*global TestTimer, JUnit, TestCase, randomInIndex, Html, print */

/**
 * A test suite which can have test cases added and run
 * Outputs HTML when done (hey that rhymed)
 * @param   {String}   name        The test suite's name
 * @param   {String}   description The test suite's description
 */
function TestSuite(name, description) {
    var suiteDescription = null;            // The text used for the description of the suite
    
    if (typeof description !== 'undefined') {
        suiteDescription = description;
    }

    
    // Since the suite can estimate the time remaining, these could be switched out for durations rather
    // than iterations if the logic were to be changed
    var WARN_ITERATIONS = 100000;       // How many iterations should be in a test before a warning is printed
    var ERROR_ITERATIONS = 1500000;     // How many iterations should be in a test before an exception is thrown
    
    var suiteName = name;               // The test suite's name
    var setUp = function () {};         // A function to run before every test case
    var tearDown = function () {};      // A function to run after every test case
    var junit = new JUnit();            // A JUnit object which logs XML data about this test suite
    var current = null;                 // The current test case
    var result = null;                  // The current test case's result
    var vars = [];                      // The current test case's arguments
    var expected = [];                  // The current test case's expected value(s)
    var comp = 0;                       // The current test case's comparison function                    
    var iteration = 0;                  // The number of times the current test case has run so far
    var forEachTests = [];              // A list of test cases with one or more lists of arguments to iterate through exhaustively
    var randomTests = [];               // A list of test cases with one or more lists of arguments to randomly select from
    var testCaseTimer =new TestTimer(); // A Timer object to be used to time test cases
    var testCaseIterations = 0;         // Number of times the test case ran
    var tests = [];                     // A list of test cases for the current suite
    var casePassed = true;              // Whether the current test case passed
    var numTests = 0;                   // The number of tests in this suite
    var m_logger = new Html();          // The HTML logger to use for this test suite
    this.logger = m_logger;             // The public variable to access this suite's logger

    /**
     * Function to set the current test suite's setUp function
     * @param  {Function} func - The function to set setUp to
     */
    this.setSetUp = function (func) {
        setUp = func;
    };

    /**
     * Function to set the current test suite's tearDown function
     * @param  {Function} func - The function to set tearDown to
     */
    this.setTearDown = function (func) {
        tearDown = func;
    };

    /**
     * Function to build a string of useful information to output when a test fails
     * @return the string containing the information
     */
    var errorString = function () {
        var returnString = '<b>Expected:</b>   \'' + expected + '\'\n';
        returnString += '<b>Comparison:</b> \'' + comp.name + '\'\n';
        returnString += '<b>Result:</b>     \'' + result + '\'\n';
        returnString += '<b>Variables:</b>  \'' + vars + '\'';

        return returnString;
    };
    
    /**
     * Function to perform a test's setUp, run the test, run its tearDown, and determine the result of the test and output the result
     */
    var performTest = function () {
        result = undefined;
        try {
            setUp();
            result = current.run(vars);
            tearDown();
        }
        catch(error) {
            if(typeof result === 'undefined') {
                result = error;
            }
            print('\nTest "' + m_logger.currentCase.name + '" Threw an exception:\n\t' + error + '\n');
        }
        
        expected = current.getExpected();
        if (expected.constructor === Array) {
            expected = expected[iteration];
        }

        if (!comp(result, expected)) {
            casePassed = false;
            m_logger.logError(errorString());
            junit.addFail(suiteName, current.getName(), 'Test failed, see HTML output for debug info');
        } else {
            // Log test pass in JUnit
            junit.addPass(suiteName, current.getName());
        }

    };

    
    
    /**
     * Stops the test case timer, ends the logger test case, ends the JUnit suite
     */
    var endTest = function() {
        testCaseTimer.stopTimer();

        m_logger.endTestCase(
            testCaseTimer.startDate,
            testCaseTimer.stopDate,
            casePassed,
            testCaseTimer.time / 1000,
            testCaseIterations
        );
        junit.endSuite();
    };
    
    /**
     * Returns a string of number n padded to x columns
     * @param   {Number} n The number to pad
     * @param   {Number} x The number of columns to pad to
     * @returns {String} n padded to x columns
     */
    var padNToX = function(n, x) {
        var nAsString = n.toString();
        if(nAsString.length >= x) {
            return nAsString;
        }
        if(x < 2) {
            throw new Error('X is ' + x + ', must be greater than 2');
        }
        var padString = '';
        for(var i = 1; i < x; i++) {
            padString += '0';
        }
        return String(padString + n).slice(-x);
    };
    
    /**
     * Returns a human-readable string representing the inputted duration
     * @param   {Number} totalMS The input duration, in milliseconds
     * @returns {String} The inputted time formatted in hours, minutes, seconds, and MS
     */
    var timeString = function(totalMS) { 
        var hours = (totalMS / 1000 / 60 / 60);
        var minutes = (hours - Math.floor(hours)) * 60;
        var seconds = (minutes - Math.floor(minutes)) * 60;
        var MS = (seconds - Math.floor(seconds)) * 1000;
        
        var returnString =
            padNToX(Math.floor(hours), 2) +   ':' +
            padNToX(Math.floor(minutes), 2) + ':' +
            padNToX(Math.floor(seconds), 2) + '.' +
            padNToX(Math.round(MS), 3);
        
        return returnString;
    };
    
    /**
     * Prints the estimated remaining time in the current test case
     * @param {Number} iterationsSoFar The number of iterations completed so far for this case
     * @param {Object} currentTime     The current time
     */
    var printTimeLeft = function(iterationsSoFar, currentTime) {
        var doneSoFar = (iterationsSoFar / testCaseIterations); // The ratio of completed test case iterations to the total number of them
        var timeSoFar = currentTime - testCaseTimer.startDate;  // The time taken so far for this test case

        var estimatedTotalTime = timeSoFar * (1 / doneSoFar);   // The estimated total time to complete this test case
        
        var estimatedTimeRemaining = (estimatedTotalTime - timeSoFar); // The estimated time left in this test case
        
        // Print the info we just obtained
        print((doneSoFar * 100).toFixed(2) + '% of iterations completed so far, estimated remaining time for this test: ' +
              timeString(estimatedTimeRemaining) + '\n');
    };
    
    /**
     * Method to run a test exhaustively. Takes a variable number of arguments,
     * which correspond to arguments for the testing function, but are passed in
     * as lists of all possible values for each of those arguments. It runs the
     * test for each combination of values. Gets expensive very quickly.
     * @param {arguments} arg1, arg2, arg3, ... - The arguments lists
     */
    var runTestsForAll = function () {
        var testIndexes = [];
        var listsList = [];
        var finalIndexes = [];
        var numArgs = arguments.length;
        
        testCaseIterations = 0;
        
        this.setupArrayToPass = function () {
            var arrayToPass = [];
            for (var i = 0; i < numArgs; i++) {
                arrayToPass[i] = listsList[i][testIndexes[i]];
            }
            return arrayToPass;
        };

        for (var i = 0; i < numArgs; i++) {
            testIndexes.push(0);
            var currList = arguments[i];
            listsList.push(currList);
            finalIndexes.push(currList.length - 1);
        }

        if(numArgs > 0) {
            testCaseIterations = (finalIndexes[0] + 1);
            for (i = 1; i < numArgs; i++) {
                testCaseIterations *= (finalIndexes[i] + 1);
            }
        } else {
            testCaseIterations = 1;
        }
        
        if(testCaseIterations >= ERROR_ITERATIONS) {
            throw new Error('A test case was started with ' + testCaseIterations + ' iterations. This will take too long.');
        } else if(testCaseIterations >= WARN_ITERATIONS) {
            print('WARNING: A test case was started with ' + testCaseIterations + ' iterations. This could take ages.');
        }

        // Now to perform the tests

        for (var t = 0; t < tests.length; t++) {
            casePassed = true;
            iteration = 0;
            testIndexes = [];
            for (i = 0; i < numArgs; i++) {
                testIndexes.push(0);
            }
            var varsArray = [];
            // Get current test
            current = tests[t];

            if (current.constructor !== TestCase) {
                throw new Error('An incorrect object type was added as a test case to this suite');
            }

            m_logger.startTestCase(current.getName(), current.getDescription(), testCaseIterations);
            junit.startSuite(current.getName(), tests.length);
            // Set variables and expected
            var testVars = current.getVariables();
            expected = current.getExpected(); // will always be a single variable
            comp = current.getComparison();

            testCaseTimer.startTimer();
            varsArray = this.setupArrayToPass();
            vars = varsArray.concat(testVars);

            if (expected.constructor === Array) {
                expected = expected[iteration];
            }
            
            var prevTime = testCaseTimer.startDate;
            var nextTime;

            performTest();

            for (i = 1; i < testCaseIterations; i++) {
                for (var j = 0; j < numArgs; j++) {
                    if (j === 0) {
                        testIndexes[j] += 1;
                    }
                    if (testIndexes[j] > finalIndexes[j]) {
                        testIndexes[j] = 0;
                        testIndexes[j + 1] += 1;
                    } else {
                        break;
                    }
                }
                varsArray = this.setupArrayToPass();
                vars = varsArray.concat(testVars);
                expected = current.getExpected();
                if (expected.constructor === Array) {
                    iteration++;
                    expected = expected[iteration];
                }
                performTest();
                
                nextTime = new Date();
                if(nextTime - prevTime > 5000) { // 5 seconds
                    printTimeLeft(i, nextTime);
                    prevTime = nextTime;
                    nextTime = new Date();
                }
                
            }
            endTest();
        }
    };

    /**
     * Method to run a test with arguments randomly selected from lists passed in
     * to the function. It is passed a variable number of arguments, each one a
     * list of all possible values for the corresponding argument in the test
     * function. It then randomly selects a value from each list for each argument
     * and runs the test or tests with them. The first argument to the function is
     * the number of times the test should be run. Each time the test is run the values
     * will be randomly selected again.
     * @param {Number} arguments[0] - The number of times to run each test
     * @param {Arguments} arg1, arg2, arg3, ... - The arguments lists
     */
    var randomlyTest = function () {
        var args = Array.prototype.slice.call(arguments);
        testCaseIterations = args.shift();
        
        if(testCaseIterations >= ERROR_ITERATIONS) {
            throw new Error('A test case was started with ' + testCaseIterations + ' iterations. This will take too long.');
        } else if(testCaseIterations >= WARN_ITERATIONS) {
            print('WARNING: A test case was started with ' + testCaseIterations + ' iterations. This could take ages.');
        }
        
        if (testCaseIterations.constructor !== Number) {
            throw new Error('This test case was added incorrectly - number of iterations was not a number');
        } else if(testCaseIterations - Math.round(testCaseIterations) !== 0 ) {
            throw new Error('This test case was added incorrectly - number of iterations was not an integer');
        }

        // Now to perform the tests

        for (var t = 0; t < tests.length; t++) {
            casePassed = true;
            // Get current test
            current = tests[t];

            if (current.constructor !== TestCase) {
                throw new Error('An incorrect object type was added as a test case to this suite');
            }

            m_logger.startTestCase(current.getName(), current.getDescription(), testCaseIterations);
            junit.startSuite(current.getName(), tests.length);

            // Set variables and expected
            var testVars = current.getVariables();
            expected = current.getExpected(); // will always be a single variable
            comp = current.getComparison();

            testCaseTimer.startTimer();
            var prevTime = testCaseTimer.startDate;
            var nextTime;
            
            for (var i = 0; i < testCaseIterations; i++) {
                vars = [];
                for (var j = 0; j < args.length; j++) {
                    vars.push(randomInIndex(args[j]));
                }
                vars = vars.concat(testVars);
                performTest();
                
                nextTime = new Date();
                if(nextTime - prevTime > 5000) { // 5 seconds
                    printTimeLeft(i, nextTime);
                    prevTime = nextTime;
                    nextTime = new Date();
                }
                
            }
            endTest();
        }
    };

    /**
     * Adds a test case which will be run for each item in each list of arguments
     * @param {Object} An array of test cases or a single test case
     * @param {Array} A list of possible values for the first argument
     * @param {Array} A list of possible values for the second argument, and so on...
     */
    this.addTest = function () {
        var args = Array.prototype.slice.call(arguments);
        if (args.length < 1) {
            throw new Error('This function requires at least one argument');
        }
        
        var test = args.shift();
        for (var i = 0; i < args.length; i++) {
            var currArg = args[i];
            if (currArg.constructor !== Array) {
                throw new Error('Arguments for addTest must be arrays. Argument ' + i + ' (' + currArg + ') is not an Array');
            }
        }
        if (test.constructor === Array) {
            numTests += test.length;
        } else {
            numTests += 1;
        }
        forEachTests.push([test, args]);
    };
    
     var runForEachTests = function () {
        for (var i = 0; i < forEachTests.length; i++) {
            tests = [];
            var thisTest = forEachTests[i][0];
            var varsList = forEachTests[i][1];
            if (thisTest.constructor === Array) {
                tests = thisTest;
            } else {
                tests.push(thisTest);
            }
            runTestsForAll.apply(this, varsList);
        }
    };

    var runRandomTests = function () {
        for (var i = 0; i < randomTests.length; i++) {
            tests = [];
            var thisTest = randomTests[i][0];
            var varsList = randomTests[i][1];
            if (thisTest.constructor === Array) {
                tests = thisTest;
            } else {
                tests.push(thisTest);
            }
            randomlyTest.apply(this, varsList);
        }
    };

    /**
     * Adds a test case which will be run a specified number of times with any number of randomly selected arguments
     * @param {Object} An array of test cases or a single test case
     * @param {Number} The number of times to run the random test
     * @param {Array} A list of possible values for the first argument
     * @param {Array} A list of possible values for the second argument, and so on...
     */
    this.addRandomTest = function () {
        if (arguments.length < 3) {
            throw new Error('This function requires at least three arguments');
        }
        var args = Array.prototype.slice.call(arguments);
        var test = args.shift();
        if (args[0].constructor !== Number) {
            throw new Error('Random Tests must have their number of iterations defined');
        }

        for (var i = 1; i < args.length; i++) {
            var currArg = args[i];
            if (currArg.constructor !== Array) {
                throw new Error('Arguments for addRandomTest must be arrays. Argument ' + i + ' (' + currArg + ') is not an Array');
            }
        }
        
        if(test.constructor === Array) {
            numTests += test.length;
        } else {
            numTests += 1;
        }
        
        randomTests.push([test, args]);
    };

    /**
     * Starts running the current test suite
     */
    this.runSuite = function () {
        if ( (forEachTests.length < 1) && (randomTests.length < 1) ) {
            throw new Error('Suite "' + suiteName + '" was run without any tests added');
        }
        
        m_logger.startSuite(suiteName, suiteDescription, numTests);

        if (forEachTests.length > 0) {
            runForEachTests();
        }

        if (randomTests.length > 0) {
            runRandomTests();
        }

        m_logger.endSuite();

    };
}
