/* globals EllyModel, ScriptHelper, VariableController, GPVariableManager */


/**
 * Sets selected items Variable ID
 * @param {number} id   Holds the id number
 */
//EllyModel.protoType.setVariableID = function(id) {
//    
//};

/**
 * Create a variable and add it to the list of variables
 * @param   {String}  name        The user defined name
 * @param   {String}  description The user defined description of the variable
 * @param   {String}  field       The starting datasource field to use
 * @param   {Number}  index       The starting datasource index to use
 * @param   {Boolean} wrapAtIndex Wrap around at max index
 * @returns {Number}  The newly created variable's id
 */
EllyModel.prototype.createVariable = function(name, description, field, index, wrapAtIndex) {
    if(typeof (GPVariableManager.variableIDNameMap())[name] !== 'undefined') {
        throw new Error('Variable named "' + name + '" already exists');
    }

    var flags = 0;
    if(wrapAtIndex === true) {
        flags = 1;
    }

    var newID = GPVariableManager.createVariable(); //name, description, field, index, flags
    return ScriptHelper.getVariableByID(GPVariableManager, newID);
};

EllyModel.prototype.getVariableByID = function(id) {
    return ScriptHelper.getVariableByID(GPVariableManager, id);
};

/**
 * Create a datasource and add it to the list of datasources
 * @param   {String} type The type of datasource to create
 * @param   {String} name The user defined name to use
 * @returns {Number} The newly created datasource id on success, else -1
 */
EllyModel.prototype.createDatasource = function(type, name) {
    var dataTypes = VariableController.getDatasourceTypes();
    var typeID = this.arraySearch(dataTypes,type);
    if(typeof typeID === 'undefined') {
        throw new Error('Data source type "' + type + '" is not defined');
    }
    if(typeof GPVariableManager.datasourceIDNameMap()[name] !== 'undefined') {
        throw new Error('Data source named "' + name + '" already exists');
    }

    return VariableController.createDatasource(type);
};

/**
 * Delete a datasource from the list of datasources
 * @param   {String}  sourceName The datasource name to delete
 * @returns {Boolean} True on success, else false
 */
EllyModel.prototype.deleteDatasource = function(sourceID) {
    return VariableController.deleteDatasource(sourceID);
};

/**
 * Delete a variable from the list of variables
 * @param   {String}  variableName The variable name to delete
 * @returns {Boolean} True on success, else false
 */
EllyModel.prototype.deleteVariable = function(variableID) {
    return GPVariableManager.deleteVariable(variableID);
};

/**
 * Deletes all datasources in this model
 */
EllyModel.prototype.deleteAllDatasources = function() {
    var datasourceMap = GPVariableManager.datasourceIDNameMap();
    var sources = Object.keys(datasourceMap);
    for(var i = 0; i < sources.length; i++) {
        var ID = sources[i];
        VariableController.deleteDatasource(ID);
    }
};

/**
 * Deletes all variables in this model
 */
EllyModel.prototype.deleteAllVariables = function() {
    var variableMap = GPVariableManager.variableIDNameMap();
    var sources = Object.keys(variableMap);
    for(var i = 0; i < sources.length; i++) {
        var ID = sources[i];
        GPVariableManager.deleteVariable(ID);
    }
};

/**
 * Attach a datasource to a variable
 * @param   {String}  sourceName   The id of the datasource to connect
 * @param   {String}  variableName The id of the variable to connect
 * @returns {Boolean} True on success, else false
 */
EllyModel.prototype.attachDatasourceToVariable = function(datasourceID, variableName) {
    ScriptHelper.attachVariable(GPVariableManager, variableName, datasourceID);
    //return GPCanvasModel.attachDatasourceToVariable(datasourceNameToID(sourceName), variableNameToID(variableName));
};

/**
 * Returns a list of all datasource types
 * @returns {Array} An array of the names of datasource types
 */
EllyModel.prototype.DatasourceTypes = function() {
    return VariableController.getDatasourceTypes();
};

/**
 * Returns a mapping of datasoruce names to IDs
 * @returns {Object} Map of datasoruce names to IDs
 */
EllyModel.prototype.DatasourceNamesToIDs = function() {
    var test = GPVariableManager.datasourceIDNameMap();
    return test;
};

/**
 * Returns a mapping of variable names to IDs
 * @returns {Object} Map of variable names to IDs
 */
EllyModel.prototype.VariableNamesToIDs = function() {
    return GPVariableManager.variableIDNameMap();
};

/**
 * Returns the ID of the given variable name. Throws an error if not found.
 * @param   {String} variableName The variable's name
 * @returns {Number} The ID of the variable
 */
var variableNameToID = function(variableName) {
    var variableId = GPVariableManager.variableIDNameMap()[variableName];
    if(typeof variableId === 'undefined') {
        return -1;//throw new Error('Variable "' + variableName + '" is not defined');
    }
    return variableId;
};

/**
 * Returns the ID of the given datasource name. Throws an error if not found.
 * @param   {String} datasourceName The datasource's name
 * @returns {Number} The ID of the datasource
 */
var datasourceNameToID = function(datasourceName) {
    var datasourceMap = GPVariableManager.datasourceIDNameMap();
    var datasourceID = datasourceMap[datasourceName];
    for (var key in datasourceMap) {
        if (datasourceMap[key] == datasourceName) {
            datasourceID = key;
        }
    }
    
    if(typeof datasourceID === 'undefined') {
        return -1;//throw new Error('Data source "' + datasourceName + '" is not defined');
    }
    return datasourceID;
};

/**
 * Attach a variable to all selected canvas items
 * @param   {Number}  variableId The id of the variable to use for attaching
 * @returns {Boolean} True on success, else false
 */
EllyModel.prototype.attachVariableToSelectedItems = function(variableID) {
    VariableController.setSelectedItemsVariableID(variableID);
    
//    VariableController.setSelectedItemsDatasourceID(variableID);
    //ScriptHelper.attachVariable(GPVariableManager, variableName, sourceID);
    //return VariableController.attachVariableToSelectedItems(variableNameToID(variableName));
};

/**
 * Detach all variables from all selected canvas items
 */
EllyModel.prototype.detachVariablesFromSelectedItems = function() {
    VariableController.disconnectSelectedItems();
};

/**
 * Set the index of all variables in m_variables
 * @param {Number} index The index to use
 */
EllyModel.prototype.setIndexAllVariables = function(index) {
    VariableController.setVariablesIndex(index);
};

/**
 * Returns a list of the given datasource's fields
 * @param   {String} sourceName The name of the datasource
 * @returns {Array}  A list of that datasource's fields
 */
EllyModel.prototype.getDatasourceFields = function(sourceID) {
    return ScriptHelper.getDatasourceFieldsByID(GPVariableManager, sourceID);
};

/**
 * Returns a datasource's data
 * @param   {Number} sourceName The datasource's name
 * @param   {String} field      The field in the datasource
 * @param   {Number} index      The index in the datasource
 * @returns {Object} The data in the given datasource at the given field and index
 */
EllyModel.prototype.getDatasourceDataByID = function(sourceID, field, index) {
    return ScriptHelper.getDatasourceDataByID(GPVariableManager, sourceID, field, index);
};

/**
 * Sets the data for a datasource
 * @param {String} sourceName The datasource's name
 * @param {Object} data       The data to send
 * @param {String} field      The field in the datasource
 * @param {Number} index      The index in the datasource
 */
EllyModel.prototype.setDatasourceData = function(sourceID, data, field, index) {
    var tempDataSource = ScriptHelper.getDatasourceByID(GPVariableManager, sourceID);
    tempDataSource.setData(data, field, index);
};

/**
 * Returns the maximum value of a datasource
 * @param   {String} sourceName The datasource's name
 * @param   {String} field      The name of the field
 * @returns {Number} The max index for the given field
 */
EllyModel.prototype.getDatasourceMax = function(sourceID, field) {
    return ScriptHelper.getDatasourceMaxByID(GPVariableManager, sourceID, field);
};

/**
 * Gets the value of a property from a datasourcxe
 * @param   {String} sourceName   The datasource's name
 * @param   {String} propertyName The property int he datasource
 * @returns {Object} The value of the given property in the given datasource
 */
EllyModel.prototype.getDatasourceProperty = function(sourceID, propertyName) {
    return ScriptHelper.getDatasourcePropertyByID(GPVariableManager, sourceID, propertyName);
};

/**
 * Sets the value of a property in a datasource
 * @param {String} sourceName    The datasource's name
 * @param {String} propertyName  The property's name
 * @param {Object} propertyValue The value to assign that property to
 */
EllyModel.prototype.setDatasourceProperty = function(sourceID, propertyName, propertyValue) {
    ScriptHelper.setDatasourcePropertyByID(GPVariableManager, sourceID, propertyName, propertyValue);
};

/**
 * Returns the type of a given datasource
 * @param   {String} sourceName The name of the datasource
 * @returns {String} The datasource's type
 */
EllyModel.prototype.getDatasourceType = function(sourceID) {
    return ScriptHelper.getDatasourceTypeByID(GPVariableManager, sourceID);
};

/**
 * Returns a lis tof properties for the given datasource
 * @param   {String} sourceName The datasource's name
 * @returns {Array}  An array of the datasource's properties
 */
EllyModel.prototype.getDatasourceProperties = function(sourceID) {
    var properties = ScriptHelper.getDatasourcePropertiesByID(GPVariableManager, sourceID);
    return properties;
};

/**
 * Adds a blank row to the given stringlist datasource
 * @param {String} sourceName The datasource's name
 */
EllyModel.prototype.addRowToStringListDatasource = function(sourceID) {
    var tempDataSource = ScriptHelper.getDatasourceByID(GPVariableManager, sourceID);
    tempDataSource.appendNewRow();
};

/**
 * removes a row from the given stringlist datasource
 * @param {String} sourceName The datasource's name
 * @param {Number} row        The index of the row
 */
EllyModel.prototype.removeRowFromStringListDatasource = function(sourceID, row) {
    var tempDataSource = ScriptHelper.getDatasourceByID(GPVariableManager, sourceID);
    tempDataSource.removeRow(row);
};

/**
 * Sets the model of a function datasource
 * @param {String} sourceName The datasource's name
 * @param {Object} model      A GPCanvasModel object
 */
EllyModel.prototype.setFunctionDatasourceModel = function(sourceID, model) {
    var tempDataSource = ScriptHelper.getDatasourceByID(GPVariableManager, sourceID);
    tempDataSource.setModel(model);
};

/**
 * Updates the variable properties based on variable ID
 * @param {int} variableID The ID number of the variable
 */
EllyModel.prototype.variablePropUpdate = function(variableID) {
    VariableController.variablePropertiesUpdate(variableID);  
};

/** 
 * Sets the Variable name based on variable ID
 * @param {int}    Variable ID
 * @param {String} Variable Name 
*/
EllyModel.prototype.setVariableName = function(variableID, variableName){
    VariableController.setVariableName(variableID, variableName);
};

EllyModel.prototype.getDatasourceData = function(sourceID, field, index) {
    return ScriptHelper.getDatasourceData(GPVariableManager, sourceID, field, index);
};